#include "Morse.h"
#include "MorseCodes.h"

#include "Arduino.h"
#include <EEPROM.h>

char buffer[16];

#define DIT_DAH_CACHE_SIZE 8
byte dit_ptr;
unsigned int dit_times[DIT_DAH_CACHE_SIZE];
byte dah_ptr;
unsigned int dah_times[DIT_DAH_CACHE_SIZE];
char last_char;

#undef DEBUG

#define MORSE_KEY_PIN 18
#define LEARN_PIN 19
#define BUZZER_PIN 17

// Base unit of time, a dit
#define DIT_TIME _ditTime
// A Dah is 3 dits
#define DAH_TIME _dahTime
// Gap between dits and dahs = 1 dit
#define SYMBOL_GAP_TIME _ditTime
// Gap between letters, one dah
#define LETTER_GAP_TIME _ditTime * 3
// Gap between words, 7 dits
#define WORD_GAP_TIME _ditTime * 7

#define MODE_NORMAL 0
#define MODE_LEARNING 1
#define MODE_MENU 2

#define CASE_AUTO 0
#define CASE_UPPER 1
#define CASE_LOWER 2

unsigned long timer1_millis;

Morse::Morse(Print * printr) {
  printer = printr;
  pinMode(MORSE_KEY_PIN, INPUT);
  pinMode(LEARN_PIN, INPUT); // get speed pin
  digitalWrite(LEARN_PIN, HIGH);
  pinMode(BUZZER_PIN, OUTPUT);
  _lastLearningMode = HIGH;
  resetMap();
  _caseMode = CASE_AUTO;
  _ditTime = 0;
  _dahTime = 0;
  _useTone = 1;
  timer1_millis = 0;
  last_char = 0;
  next_space = 0;
  if (LoadState() == 0) {
    startLearningMode(1);
  } else {
    _mode = MODE_NORMAL;
  }
#if defined(__AVR_ATmega8__)
    TIMSK&=!(1<<TOIE0);
    //TIMSK|=(1<<TOV1);
# else
    TIMSK0&=!(1<<TOIE0);
    TIMSK1|=(1<<TOIE1);
#endif
}

  ISR(TIMER1_OVF_vect) {
    sei();
    timer1_millis += 2;
  }


  long my_millis() {
    unsigned long m;
    uint8_t oldSREG = SREG;

    cli();
    m = timer1_millis;
    SREG = oldSREG;
    sei();
    return m;
  }

void Morse::resetMap() {
  _keyLastEvent = my_millis();
  _keyLastPos = LOW;
  _keyTimePos = 0;
  memset(_keyTimeMap, 0, sizeof(_keyTimeMap));
}

void Morse::SaveState() {
  EEPROM.write(0, 0x69); // Magic number

  EEPROM.write(1, (_ditTime & 0xff));
  EEPROM.write(2, ((_ditTime >> 8) & 0xff));

  EEPROM.write(3, (_dahTime & 0xff));
  EEPROM.write(4, ((_dahTime >> 8) & 0xff));

  EEPROM.write(5, _caseMode);
  EEPROM.write(6, _useTone);
}

char Morse::LoadState() {
  if (EEPROM.read(0) != 0x69)
    return 0; // invalid magic

  _ditTime = EEPROM.read(2) << 8 | EEPROM.read(1);
  _dahTime = EEPROM.read(4) << 8 | EEPROM.read(3);
  _caseMode = EEPROM.read(5);
  _useTone = EEPROM.read(6);
  return 1;
}

char Morse::startLearningMode(char force) {
  if (force || (_lastLearningMode == LOW && digitalRead(LEARN_PIN))) { // Learning mode
    if (!force)
      printer->println("\nLearning. Enter '-.-.-'");
    _mode = MODE_LEARNING;
    _ditTime = 5000;
    _lastLearningMode = HIGH;
    resetMap();
    return 1;
  }
  _lastLearningMode = digitalRead(LEARN_PIN);
  return 0;
}

unsigned char Morse::getLetterFromCode(const char *code) {
char buffer[16];
char code_num;
char found = 0;
  for (code_num = 0 ; code_num < MORSE_CHAR_COUNT ; code_num++) {
    strcpy_P(&buffer[0], (char *)pgm_read_word(&(morse_table[code_num])));
    if (!strcmp(code, &buffer[0])) {
      found = 1;
      break;
    }
  }
#if 0
  printer->print(code);
  printer->print(" is ");
  printer->print(code_num, DEC);
  printer->print(", ");
  printer->println(pgm_read_byte_near(morse_chars + code_num));
#endif
  if (!found)
    return '?';
  return pgm_read_byte_near(morse_chars + code_num);
}

unsigned int average(unsigned int values[]) {
unsigned int total = 0;
char nvalues = 0;
  for (char i = 0 ; i < DIT_DAH_CACHE_SIZE ; i++)
    if (values[i] != 0) {
      total += values[i];
      nvalues++;
    }
  return total / nvalues;
}

unsigned char Morse::decodeSymbol() {
  // Decode symbol in the keyMap buffer.
  char morse[16];
  unsigned char keycode;
  unsigned char morsepos = 0;
  unsigned int keyEvent;
  memset(&morse[0], 0, 16);
  for (char i = 0 ; i < _keyTimePos ; i++) {
    keyEvent = _keyTimeMap[i];
    if (keyEvent & 0x1) continue;
    if (keyEvent < 1.5*DIT_TIME) {
      morse[morsepos++] = '.';
      dit_times[dit_ptr++] = keyEvent;
      dit_ptr &= DIT_DAH_CACHE_SIZE-1;
    } else if (keyEvent < 1.5*DAH_TIME) {
      morse[morsepos++] = '-';
      dah_times[dah_ptr++] = keyEvent;
      dah_ptr &= DIT_DAH_CACHE_SIZE-1;
    }
  }
  keycode = getLetterFromCode(&morse[0]);
/*
  if (keycode == '?') {
    printer->println("DITS\n");
    printer->print("AVG: ");
    printer->print(average(dit_times));
    printer->println();
    printer->println("DAHS\n");
    printer->print("AVG: ");
    printer->print(average(dah_times));
    printer->println();
  }
*/
  _ditTime = average(dit_times);
  _dahTime = average(dah_times);
//  if (_caseNext == CASE_LOWER && keycode < 91 && keycode > 64)
//    return keycode + 32;
  return keycode;
}

uint8_t Morse::getWpm() {
byte wpm;
  wpm = 1200 / _ditTime;
  //return 0x57;

  return ((wpm/10) << 4) + (wpm % 10);
}

void Morse::calculateTiming() {
// Work out timing from the start symbol. (_._._)
  dit_ptr = 0;
  dah_ptr = 0;
  for (char i = 0 ; i <= _keyTimePos ; i++) {
    if (_keyTimeMap[i] & 0x1)
      continue;
    if (i == 3 || i == 7)
      dit_times[dit_ptr++] = _keyTimeMap[i];
    else
      dah_times[dah_ptr++] = _keyTimeMap[i];
  }
  printer->print("Avg dit (ms): ");
  _ditTime = average(dit_times);
  printer->println(_ditTime);
  printer->print("Avg dah (ms): ");
  printer->println(average(dah_times));
  _dahTime = average(dah_times);
  _mode = MODE_NORMAL;
  printer->println("Done. 'SOS' to enter main menu.");
}

void Morse::showMenu() {
  printer->println("Settings.:");
  printer->println("  1. Normal.");
  printer->println("  2. Learning.");
  printer->println("  3. Auto-format.");
  printer->println("  4. Sidetone.");
  printer->print("-> ");
  resetMap();
}

uint8_t Morse::poll() {
  // Main loop. Poll pins, etc.

  if (startLearningMode(0)) {
    return 0xAA;
  }

  char keyPos = digitalRead(MORSE_KEY_PIN);
  unsigned int duration = (int)(my_millis() - _keyLastEvent);
  if (duration < DEBOUNCE_TIME) // Not debounced
    return 0;

  if (_useTone && keyPos)
    tone(BUZZER_PIN, 500);
  else
    noTone(BUZZER_PIN);

  if (keyPos == _keyLastPos) { // No key change
    if (duration >= LETTER_GAP_TIME && !keyPos && _keyTimePos) {
      char code = decodeSymbol();
      if (_mode == MODE_NORMAL) {
        if (code == 0x7) {  // SOS, go to menu mode
          _mode = MODE_MENU;
          showMenu();
          return 0xbb;
        } else {
          // Letter gap, so print letter.
          if (_caseMode == CASE_AUTO) {
            if (last_char == '.') {
              printer->print(' ');
              if (code <= 'z' && code >= 'a')
                code -= 32;
            } else if (next_space) {
              if (last_char != '\b' && last_char != '\n' &&
                  last_char != '.') {
                switch(code) {
                  case '\b':
                  case '\n':
                  case '.':
                  case ',':
                    break;
                  default:
                    printer->print(' ');
                    SaveState();
                }
              }
            }
          }
          printer->print(code);
          next_space = 0;
          last_char = code;
          resetMap();
          return getWpm();
        }
      }
      if (_mode == MODE_MENU) {
        switch(code) {
          case '1':
            printer->println("Normal mode. 'SOS' to reenter this menu.");
            _mode = MODE_NORMAL;
            SaveState();
            resetMap();
            return getWpm();
            break;
          case '2':
            printer->println("\nLearning. Enter '-.-.-'");
            startLearningMode(1);
            return 0xAA;
            break;
          case '3':
            _caseMode = (_caseMode == CASE_AUTO ? CASE_LOWER : CASE_AUTO);
            printer->print("Auto format ");
            printer->println(_caseMode == CASE_AUTO ? "ON" : "OFF");
            printer->print("-> ");
            break;
          case '4':
            _useTone = 1 - _useTone;
            printer->print("Sidetone ");
            printer->println(_useTone ? "ON" : "OFF");
            printer->print("-> ");
            break;
          default:
            printer->println("Unknown code. Try again. ");
            printer->print("-> ");
            break;
        }
        SaveState();
      }
      resetMap();
    }
    return 0;
  }

  // Key changed from here...

  if (keyPos == HIGH && duration > 2*_dahTime)
    next_space = 1;

  _keyLastEvent = my_millis();
  _keyLastPos = keyPos;

  // Store duration. LSB is whether its an up or down stroke.
  _keyTimeMap[_keyTimePos] = (duration & 0xFFFE) + keyPos;

#ifdef DEBUG
  if (keyPos && duration > LETTER_GAP_TIME) // Letter gap, print newline
    printer->println();
  else if (_keyTimePos) { // Show graph of current symbol
    for (char i = 0 ; i < duration/30 ; i++)
      printer->print(keyPos ? '_' : '|');
  }
#endif

  if (_mode == MODE_LEARNING && _keyTimePos > 8) {
    calculateTiming();
    resetMap();
    return getWpm();
  }

  _keyTimePos ++;
  //_keyTimePos &= KEYTIME_POS_SIZE - 1;
  if (_keyTimePos > 20) _keyTimePos = 0;
  return 0;
}

char *Morse::getCode(char letter) {
  char code_num;
  memset(buffer, 0, sizeof(buffer));
  for (code_num = 0 ; code_num < MORSE_CHAR_COUNT ; code_num++)
    if (pgm_read_byte_near(morse_chars + code_num) == letter)
      strcpy_P(buffer, (char *)pgm_read_word(&(morse_table[code_num])));
  return &buffer[0];
}
